/*============================================================================

   Name     : StartEndFactory.java

   System   : FME Plug-in Builder Sample

   Language : Java

   Purpose  : Implementation of an FME factory called 'StartEndFactory'

   Author               Date            Changes made
   ------------------   ------------    -------------------------------
   Your Name            Xxx XX, 200X    Original definition

                  Copyright (c) 1999 - 2000, Your Company Goes Here
                              All Rights Reserved

    The entire risk as to the results and performance of the software,
    supporting text and other information contained in this file
    (collectively called the "Software") is with the user.  Although
    Safe Software Incorporated has used considerable efforts in preparing
    the Software, Safe Software Incorporated does not warrant the
    accuracy or completeness of the Software. In no event will Safe Software
    Incorporated be liable for damages, including loss of profits or
    consequential damages, arising out of the use of the Software.

    NOTE: Replace the above header with whatever is appropriate for your
    company.

============================================================================*/

package com.mycompany.myfactory;

import COM.safe.fme.pluginbuilder.*;
import COM.safe.fmeobjects.*;
//import java.io.*;
import java.util.*;

public final class StartEndFactory implements IFMEFactory
{

   //===========================================================================
   // Constants for this factory:

   // Keywords
   private final String kAlgorithm = "ALGORITHM"; // this is just to show how to use
                                                  // keywords

   // Output tags
   private final String kStart = "START";
   private final String kEnd = "END";
   private final String kOriginal = "ORIGINAL";

   // private members
   private IFMEMappingFile mappingFile_ = null;
   private IFMELogFile logFile_ = null;
   private IFMEFactoryBridge bridge_ = null;
   private IFMECoordSysManager coordSysMan_ = null;
   private IFMESession session_ = null;

   private int currentSourceDimension_ = IFMEFeature.FME_TWO_D;
   private static final int kDefaultNumCoords = 10;

   //===========================================================================
   // constructor
   //
   StartEndFactory(IFMEFactoryBridge bridge) throws Exception
   {
      bridge_ = bridge;
   }

   //===========================================================================
   // In the initialize method we have to set up the keywords we are interested
   // in AND the output tags we want.  For this factory, we have no input tags.
   //
   public void initialize(IFMEMappingFile mappingFile,
                          IFMELogFile logFile,
                          IFMECoordSysManager coordSysMan,
                          IFMESession session)
                          throws Exception
   {

      //In this example we have a keyword but we ignore it.
      //A keyword is a word that appears in the factory definition, but FME
      //doesn't recognize it and doesn't know what to do with it, thus, we must
      //define and set the keyword ourselves.
      //EX:   FACTORY_DEF * StartEndFactory   \
      //  INPUT FEATURE_TYPE *                \ FME knows how to handle INPUT statement
      //  ALGORITHM plane_sweep               \ FME doesn't know how to handle ALGORITHM statement
      //  OUTPUT END FEATURE_TYPE END           FME knows how to handle OUTPUT statement


      mappingFile_ = mappingFile;
      logFile_ = logFile;
      coordSysMan_ = coordSysMan;
      session_ = session;

      ArrayList keywords = new ArrayList();
      keywords.add(kAlgorithm);

      bridge_.setKeywords(keywords);

      // Now the output tags
      ArrayList outputTags = new ArrayList();
      outputTags.add(kStart);
      outputTags.add(kEnd);
      outputTags.add(kOriginal);
      bridge_.setOutputTags(outputTags);

      // And that's it -- no input tags to care about

      // Prove that we've been here - log a message.
      logFile_.logMessageString("Inside the 'initialize' method.",
                                IFMELogFile.FME_INFORM);
   }

   //===========================================================================
   // This is called when the keyword is found in the FACTORY_DEF
   //
   public void processClause(ArrayList factoryClause) throws Exception
   {
      // Okay, just to convince ourselves that we got here, we'll log the clause.
      // Other times, we may need to do something useful.
      logFile_.logMessageString("StartEndJavaFactory (inside 'processClause'): Processing StartEndJavaFactory clause: " +
                                (String)factoryClause.get(0), IFMELogFile.FME_INFORM);

      String keyword = (String)factoryClause.get(0);

      if ((keyword.equals(kAlgorithm)) && (factoryClause.size() == 2))
      {
         logFile_.logMessageString("StartEndJavaFactory: Using Algorithm " +
                                   (String)factoryClause.get(1) +
                                   " (this isn't useful, but it's just for demonstration)",
                                   IFMELogFile.FME_INFORM);
      }
      else
      {
         String message = "";
         switch (factoryClause.size())
         {
            case 5:
               message = " " + (String)factoryClause.get(4);
               // Fallthrough
            case 4:
               message = " " + (String)factoryClause.get(3) + message;
               // Fallthrough
            case 3:
               message = " " + (String)factoryClause.get(2)  +  message;
               // Fallthrough
            case 2:
               message = " " + (String)factoryClause.get(1)  + message;
               // Fallthrough
            case 1:
               message = " " + (String)factoryClause.get(0)  + message;
               // Fallthrough
            default:
               message = "Invalid specification:" + message;
         }
         throw new Exception(message);
      }
   }


   //===========================================================================
   // Depending on how we have chosen to process features in this factory we could:
   // a) Process and output each feature immediately or
   // b) Process and store features into a list, then when all features
   // have been collected, process and output them.
   // Under a) the bulk of the work occurs in this function since features are output immediately as soon as
   // they are done processing
   // Under b) the bulk of the works occurs in the close() function. Within this function
   // assuming we have created a feature vector(requires work in initialization() to create an FMEObjects
   // session and the feature vector) that can store features in a list, then all we do in this function
   // is append the feature to the feature vectore like this:
   //       featureList_->append(feature);
   // And then only in the close() function after we have collected all the features does
   // processing occur. (processing is exactly the same as what would occur here except it is
   // looped for each feature in the list until no more features remain)
   //
   public void process(IFMEFeature feature, String inputTag) throws Exception
   {
      //inputTag does not have to be looked at.
      if (feature.numCoords() > 0)
      {
         // We will create two new features out of this guy, one for the
         // start point and one for the end point  These are the
         // "NODE" features we will output.
         // Clone the attributes and feature type into the
         // point features.  Actually, the method clones
         // everything EXCEPT the coordinates themselves.

         IFMEFeature firstPoint = session_.createFeature();
         feature.clone(firstPoint);
         firstPoint.resetCoords();

         IFMEFeature lastPoint = session_.createFeature();
         feature.clone(lastPoint);
         lastPoint.resetCoords();

         // for debugging purposes
         //IFMEFeature firstPoint = bridge_.createFeature();
         //IFMEFeature lastPoint = bridge_.createFeature();

         // Now get the first and last points of the original
         // feature and set them as the only coords of our two
         // fake "NODE"s

         double [] coord = new double[feature.getDimension()];

         // Get the first coordinate and place it into the 'firstPoint'
         // feature.
         if (feature.getDimension() == 2)
         {
            coord = feature.get2DCoordinate(0);
            firstPoint.add3DCoordinate(coord[0], coord[1], 0.0);
         }
         else
         {
            coord = feature.get3DCoordinate(0);
            firstPoint.add3DCoordinate(coord[0], coord[1], coord[2]);
         }

         // Set the geometry type for the 'firstPoint' feature.
         firstPoint.setGeometryType(IFMEFeature.FME_GEOM_POINT);

         // Get the first coordinate and place it into the 'firstPoint'
         // feature.
         if (feature.getDimension() == 2)
         {
            coord = feature.get2DCoordinate(feature.numCoords() - 1);
            lastPoint.add3DCoordinate(coord[0], coord[1], 0.0);
         }
         else
         {
            coord = feature.get3DCoordinate(feature.numCoords() - 1);
            lastPoint.add3DCoordinate(coord[0], coord[1], coord[2]);
         }

         // Set the geometry for the 'lastPoint' feature.
         lastPoint.setGeometryType(IFMEFeature.FME_GEOM_POINT);

         // output the START feature (otherwise known as the 'firstPoint' feature)
         bridge_.outputFeature(kStart, firstPoint);

         // output the END feature (otherwise known as the 'lastPoint' feature)
         bridge_.outputFeature(kEnd, lastPoint);

      }

      // In any case, we send the original feature out and on its way

      // output the ORIGINAL feature
      bridge_.outputFeature(kOriginal,feature);
   }

   //===========================================================================
   // Sometimes this function is used to go through all input features but
   // in this factory we didn't need to save them.
   //
   public void close() throws Exception
   {
      // do nothing
   }

   //===========================================================================
   // abort
   //
   public void abort() throws Exception
   {
      // do nothing
   }

   //===========================================================================
   // id
   //
   public long id()
   {
      return (long)123456789;
   }
}
