/*============================================================================

   Name     : RedReader.java

   System   : FME Developer Kit

   Language : Java

   Purpose  : Implements the IFMEReader interface to supply a reader for
              the Red format (near-identical to Arc/Info Generate).

   Author               Date            Changes made
   ------------------   ------------    -------------------------------
   Your Name            Xxx XX, 200X    Original definition

                  Copyright (c) 1999 - 2000, Your Company Goes Here
                              All Rights Reserved

    The entire risk as to the results and performance of the software,
    supporting text and other information contained in this file
    (collectively called the "Software") is with the user.  Although
    Safe Software Incorporated has used considerable efforts in preparing
    the Software, Safe Software Incorporated does not warrant the
    accuracy or completeness of the Software. In no event will Safe Software
    Incorporated be liable for damages, including loss of profits or
    consequential damages, arising out of the use of the Software.

    NOTE: Replace the above header with whatever is appropriate for your
    company.

============================================================================*/

/******************************************************************************
Class Overview:
================
Implements the IFMEReader interface to supply a reader for the Red format.

With two exceptions, the Red Reader is functionally identical to the FME ARCGEN Reader
used to read ESRI's Arc/Info Generate format. The two exceptions are:

1) Red does not require nor use any source DEF lines in the mapping file.
2) Only the comma character is valid as the delimiter character in its source files.
3) The data file extension must be ".red".

Details of the ARCGEN Reader (and therefore of the Red Reader) and the Generate
format can be found in the FME Reference Manual Part 2, in the Section
"ESRI ArcInfo Generate File Reader/Writer".
******************************************************************************/


package com.mycompany.myformat.myreader;

import COM.safe.fme.pluginbuilder.*;
import COM.safe.fmeobjects.*;
import java.io.*;
import java.util.*;



//---------------------------------------------------------------------
// RedReader class.
public final class RedReader implements IFMEReader, FilenameFilter
{
   // Public constants.

   // The extension for Red source files.
   public static final String kFormatExtension = "red";

   public static final String kComma = ",";
   public static final int kDefaultNumCoords = 100;
   public static final int kLookAheadLimit = 500;

   // Private instance variables.
   private IFMEMappingFile mappingFile_ = null;
   private IFMELogFile logFile_ = null;
   private IFMECoordSysManager coordSysMan_ = null;

   private String readerType_ = null;
   private String readerKeyword_ = null;

   private String currentSourceBasename_ = null;
   private String currentSourceGeomType_ = null;

   private int numSourceFiles_ = 0;
   private int currentSourceFileNum_ = -1;
   private boolean nextSourceFileOpen_ = false;
   private boolean checkDimension_ = true;
   private int currentSourceDimension_ = IFMEFeature.FME_TWO_D;
   private File[] sourceFiles_ = new File[1];

   private LineNumberReader sourceFileReader_ = null;

   // -----------------------------------------------------------------------
   // Package constructor.
   //
   // Takes the reader type (Red) and the reader keyword.
   public RedReader(String readerType, String readerKeyword)
   {
      readerType_ = readerType;
      readerKeyword_ = readerKeyword;
   }

   // -----------------------------------------------------------------------
   // init
   //
   // Initialize the reader with the mapping file, log file, coodinate system
   // manager and session objects.
   //
   public void init(IFMEMappingFile mappingFile,
               IFMELogFile logFile,
               IFMECoordSysManager coordSysMan,
               IFMESession session)
               throws Exception
   {
      mappingFile_ = mappingFile;
      logFile_     = logFile;
      coordSysMan_ = coordSysMan;

/*  The Red Reader does not make use of source DEF lines in the mapping file.
    However, many Readers will often do so. Below is sample code that illustrates
    how the Red Reader would access and process DEF lines that define the geometry
    type and delimiter character for each source file. This functionality is in fact
    supplied by the FME ARCGEN Reader, and the sample code below assumes that the
    DEF lines being processed follow the ARCGEN syntax for source DEF lines. An
    example of such a DEF line is:

      ARCGEN_DEF roads                \
        ARCGEN_GEOMETRY arcgen_line   \
        ARCGEN_DELIMITER "|"

    In the code below, each DEF line is accessed as an array of tokens. Each
    element in the DEF line constitutes one token, so the example above would contain
    6 tokens. Note that the delimiter character token is returned with the surrounding
    double-quote characters already removed, so the token consists of just the single pipe
    character without the double quotes.

    // Create hashmaps to store the geometry type and delimiter for each source file.
    HashMap sourceFileGeomType_ = new HashMap();
    HashMap sourceFileDelimiter_ = new HashMap();

    // Get an iterator on the DEF lines returned by the mapping file object.
    Iterator defLinesIterator = mappingFile_.fetchDefLines();

    while (defLinesIterator.hasNext())
    {
      // Get the next DEF line as an array of strings, each element representing
      // a token in the DEF line.
      ArrayList defLineTokens = (ArrayList)defLinesIterator.next();

      if (defLineTokens.size() >= 4)
      {
        // If at least 4 tokens exist, assume there is a geometry specified.
        // Enter the basename as the key and the geometry string as the value
        // into the hashmap.
        sourceFileGeomType_.put(defLineTokens.elementAt(1),
                                defLineTokens.elementAt(3));

        if (defLineTokens.size() == 6)
        {
          // If 6 tokens exist, assume there is a delimiter specified.
          // Enter the basename as the key and the delimiter character string as the
          // value into the hashnap.
          sourceFileDelimiter_.put(defLineTokens.elementAt(1),
                                   defLineTokens.elementAt(5));
        }
        else
        {
          // Otherwise assume there is no explicit delimiter specification so add
          // the comma as the default delimiter.
          sourceFileDelimiter_.put(defLineTokens.elementAt(1),
                                   kComma);
        }
      }
    }
*/
   }

   // -----------------------------------------------------------------------
   // open
   //
   // Open up the reader and pass it the string array containing dataset parameters.
   // The first parameter is the source dataset directory in which the Red data files
   // reside. Subsequent parameters can include one or more occurrences of the pair
   // "+ID" followed by the basename of a Red source file to process.
   //
   // If such "+ID" pairs are present, the reader will read only these files, ignoring all
   // others in the dataset directory specified by the first parameter. If no "+ID" pairs
   // are present, the reader processes all Red files in the directory.
   //
   // For example, datasetParams could contain the following 5 elements:
   //
   //   "D:\SourceData\Municipal\Roads" "+ID" "road" "+ID" "sewer"
   //
   // The first string is the directory in which the Red source files reside. The next element
   // is a "+ID" keyword followed by the basename "road". Another "+ID"-"basename" pair follows this,
   // specifying a basename of "sewer". This indicates to the Reader that it should process only the
   // Red source files "road.red" and "sewer.red" in the indicated source dataset directory.
   //
   // If anything goes wrong an exception is thrown describing the type of error that occurred.
   public void open(ArrayList datasetParams) throws Exception
   {
//      logFile_.logMessageString("In open()",IFMELogFile.FME_INFORM);
//      for (int i=0; i<datasetParams.size(); i++)
//      {
//         logFile_.logMessageString((String)(datasetParams.get(i)),IFMELogFile.FME_INFORM);
//      }

      // Create a File representing the source dataset directory.
      File datasetDir = new File((String)(datasetParams.get(0)));

      // Determine the number of dataset parameters passed to this method.
      int numParams = datasetParams.size();

      ArrayList sourceFilesByID = new ArrayList();
      boolean idsArePresent = false;

      // Check each dataset parameter.
      for (int paramNum = 1; paramNum < numParams; ++paramNum)
      {
         if (datasetParams.get(paramNum).equals("+ID"))
         {
            // If the parameter is "+ID" there are ids present, so process only the files specified.
            idsArePresent = true;
            // Create a File representing the source file whose basename is specified by the next
            // parameter after "+ID". Add the Red format extension to it first.
            File sourceFile = new File(datasetDir, datasetParams.get(++paramNum) + "." + kFormatExtension);

            if (sourceFile.exists())
            {
               // If the specified source file actually exists on the filesystem in the source dataset
               // directory, add its File object to the list of files to process.
               sourceFilesByID.add(sourceFile);
            }
            else
            {
               // But if the file doesn't exist, log a warning and continue.
               logFile_.logMessageString("Cannot find source file '" +
                                         sourceFile.getAbsolutePath() +
                                         "'.",IFMELogFile.FME_WARN);
            }
         }
      }

      if (sourceFilesByID.size() > 0)
      {
         // If there are source files in the list, get the list as a File array and set the
         // number of source files. Getting a File array is required in order to be consistent
         // with the return type of the File listFiles() method (see below).
         sourceFiles_ = (File[])sourceFilesByID.toArray(sourceFiles_);
         numSourceFiles_ = sourceFilesByID.size();
      }
      else
      {
         // There are no source files in the list.
         if (idsArePresent)
         {
            // If ids were present, then the files they specified didn't exist, so
            // set the File array to null.
            sourceFiles_ = null;
            numSourceFiles_ = 0;
         }
         else
         {
            // But if no ids were present, then all source files should be processed.
            // So get a list of all source files in the datsaet directory, using this
            // class as the FilenameFilter to accept only files with the correct extension.
            sourceFiles_ = datasetDir.listFiles(this);
            numSourceFiles_ = sourceFiles_.length;
         }
      }

      if ((sourceFiles_ == null) || (numSourceFiles_ <= 0))
      {
         // If the source file list is null or no source files were found, throw
         // an exception.
         throw new Exception("Problem accessing source directory " +
                             datasetParams.get(0) +
                             " or finding RED source files in it.  " +
                             "Source files must be present in directory with '.red' extension and named with IDs keyword if used.");
      }

      // Otherwise, open the next (first) source file in the list and get back the
      // expected "true" return status.
      nextSourceFileOpen_ = openNextSourceFile();
   }

   // -----------------------------------------------------------------------
   // abort
   //
   // Terminate the reader mid-stream.  Any special actions to shut
   // down a reader not finished reading data should be taken in this
   // method.  For most readers, nothing will be done.
   public void abort() throws Exception
   {
   }

   // -----------------------------------------------------------------------
   // close
   //
   // Close the reader after it has exhausted its data source.
   public void close() throws Exception
   {
   }

   // -----------------------------------------------------------------------
   // id
   //
   // Return the unique ID assigned to this reader.
   public int id()
   {
      // All Java reader return 0.
      return 0;
   }

   // -----------------------------------------------------------------------
   // read
   //
   // Read the next feature in the data source, filling in the
   // empty feature object provided.  If an error occurs, an exception
   // is thrown, otherwise a returned feature object indicates that all
   // is well. If the data source is exhausted and no feature data could
   // be read, then the method returns null.
   public IFMEFeature read(IFMEFeature feature) throws Exception
   {
      String sourceLine = null;

//      logFile_.logMessageString("In read() ******",IFMELogFile.FME_INFORM);
      try
      {
         // Read in the next line from the current source file.
         sourceLine = sourceFileReader_.readLine();

         if ((sourceLine == null) || sourceLine.startsWith("END"))
         {
            // If the end of the source file has been reached, close it.
            try {sourceFileReader_.close();}
            catch (IOException e)
            {
               // Throw exception if can't close.
               throw new Exception("Problem closing source file " +
                                   sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                                   ".");
            }

            if (openNextSourceFile())
            {
               // Open the next source file in the list (if another one is available)
               // and read in the first line from it.
               sourceLine = sourceFileReader_.readLine();
            }
            else
            {
               // If all source files have been processed, return null feature.
               return null;
            }
         }
      }
      catch (IOException e)
      {
         // Throw exception if problem reading from the source file.
         throw new Exception("Problem reading source file " +
                             sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                             " at line " +
                             Integer.toString(sourceFileReader_.getLineNumber()) +
                             ".");
      }

      // Set the feature type (same as cource file's basename).
      feature.setFeatureType(currentSourceBasename_);

      if (currentSourceGeomType_.equals("red_point"))
      {
         // If source file is point geometry, set as such and read in a point feature.
         feature.setGeometryType(IFMEFeature.FME_GEOM_POINT);
         return readPointFeature(sourceLine, feature);
      }
      else if (currentSourceGeomType_.equals("red_line"))
      {
         // If source file is line geometry, set as such and read in a line feature.
         feature.setGeometryType(IFMEFeature.FME_GEOM_LINE);
         return readLineFeature(sourceLine, feature);
      }
      else if (currentSourceGeomType_.equals("red_text"))
      {
         // If source file is text geometry, set as point and read in a text feature.
         feature.setGeometryType(IFMEFeature.FME_GEOM_POINT);
         return readTextFeature(sourceLine, feature);
      }
      else
      {
         // Otherwise throw exception for unknown source geometry.
         throw new Exception("Unknown source geometry type " +
                             currentSourceGeomType_ +
                             " for source feature type " +
                             currentSourceBasename_ +
                             ".");
      }
   }

   // -----------------------------------------------------------------------
   // readPointFeature
   //
   // Read in a point feature. For point source files, each source line represents one feature.
   private IFMEFeature readPointFeature(String sourceLine, IFMEFeature feature) throws Exception
   {
      // Get tokenizer on the source line.
      StringTokenizer sourceLineTokenizer = new StringTokenizer(sourceLine, kComma);

      int idNum = 0;

      double xCoord = 0.0,
             yCoord = 0.0,
             zCoord = 0.0;

      try
      {  // Extract the id, x & y coords from the line.
         idNum = Integer.parseInt(sourceLineTokenizer.nextToken());
         xCoord = Double.valueOf(sourceLineTokenizer.nextToken()).doubleValue();
         yCoord = Double.valueOf(sourceLineTokenizer.nextToken()).doubleValue();

         if (checkDimension_)
         {
            // If the dimensionality should be checked & set for this source file...
            if (sourceLineTokenizer.hasMoreTokens())
            {
               // If there are more tokens after the y-coord, assume there is a z-ccord
               // present, so set dimensionality to 3D.
               feature.setDimension(IFMEFeature.FME_THREE_D);
               currentSourceDimension_ = IFMEFeature.FME_THREE_D;
            }
            else
            {
               // Otherwise, there's no z-ccord present, so set dimensionality to 2D.
               feature.setDimension(IFMEFeature.FME_TWO_D);
               currentSourceDimension_ = IFMEFeature.FME_TWO_D;
            }
            // Dimensionality has been checked & set for this source file, so don't
            // do it again.
            checkDimension_ = false;
         }

         if (currentSourceDimension_ == IFMEFeature.FME_THREE_D)
         {
            // If this is a 3D source file, get the z-coord.
            zCoord = Double.valueOf(sourceLineTokenizer.nextToken()).doubleValue();
         }
      }
      catch (NoSuchElementException e)
      {
         // Throw exception if there aren't enough tokens in the source line.
         throw new Exception("Not enough values in point source file " +
                             sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                             " at line " +
                             Integer.toString(sourceFileReader_.getLineNumber()) +
                             ".");
      }
      catch (NumberFormatException e)
      {
         // Throw exception if there are non-numeric values in the source line.
         throw new Exception("Non-numeric value for feature in point source file " +
                             sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                             " at line " +
                             Integer.toString(sourceFileReader_.getLineNumber()) +
                             ".");
      }
      // Otherwise, add the coords to the point feature.
      feature.add3DCoordinate(xCoord, yCoord, zCoord);
      // Set the point feature "red_id" attribute.
      feature.setIntAttribute("red_id",idNum);
      // Return the point feature.
      return feature;
   }

   // -----------------------------------------------------------------------
   // readLineFeature
   //
   // Read in a line feature. For line source files, each feature is represented by a
   // set of source lines. So reading in a line feature requires reading in more
   // than one source line, unlike point and text features.
   private IFMEFeature readLineFeature(String sourceLine, IFMEFeature feature) throws Exception
   {
      int idNum = 0;

      double xCoord = 0.0,
             yCoord = 0.0,
             zCoord = 0.0;

      try
      {
         // Get the id from the first source line.
         idNum = Integer.parseInt(sourceLine);
         // Then read in the next line.
         sourceLine = sourceFileReader_.readLine();

         // Read in source lines making up the feature until the "END" line is reached,
         // signalling the end of the current line feature.
         while (!sourceLine.startsWith("END"))
         {
            // Get tokenizer on the source line.
            StringTokenizer sourceLineTokenizer = new StringTokenizer(sourceLine, kComma);
            // Extract the id, x & y coords from the line.
            xCoord = Double.valueOf(sourceLineTokenizer.nextToken()).doubleValue();
            yCoord = Double.valueOf(sourceLineTokenizer.nextToken()).doubleValue();

            if (checkDimension_)
            {
               // If the dimensionality should be checked & set for this source file...
               if (sourceLineTokenizer.hasMoreTokens())
               {
                  // If there are more tokens after the y-coord, assume there is a z-ccord
                  // present, so set dimensionality to 3D.
                  feature.setDimension(IFMEFeature.FME_THREE_D);
                  currentSourceDimension_ = IFMEFeature.FME_THREE_D;
               }
               else
               {
                  // Otherwise, there's no z-ccord present, so set dimensionality to 2D.
                  feature.setDimension(IFMEFeature.FME_TWO_D);
                  currentSourceDimension_ = IFMEFeature.FME_TWO_D;
               }
               // Dimensionality has been checked & set for this source file, so don't
               // do it again.
               checkDimension_ = false;
            }

            if (currentSourceDimension_ == IFMEFeature.FME_THREE_D)
            {
               // If this is a 3D source file, get the z-coord.
               zCoord = Double.valueOf(sourceLineTokenizer.nextToken()).doubleValue();
            }
            // Add the coords to the line feature.
            feature.add3DCoordinate(xCoord, yCoord, zCoord);
            // Read in the next source line of the current line feature.
            sourceLine = sourceFileReader_.readLine();
         }
      }
      catch (IOException e)
      {
         // Throw exception if problem reading from the source file.
         throw new Exception("Problem reading line source file " +
                              sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                            " at line " +
                            Integer.toString(sourceFileReader_.getLineNumber()) +
                            ".");
      }
      catch (NoSuchElementException e)
      {
         // Throw exception if there aren't enough tokens in the source line.
         throw new Exception("Not enough values in line source file " +
                             sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                             " at line " +
                             Integer.toString(sourceFileReader_.getLineNumber()) +
                             ".");
      }
      catch (NumberFormatException e)
      {
         // Throw exception if there are non-numeric values in the source line.
         throw new Exception("Non-numeric value for feature in line source file " +
                             sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                             " at line " +
                             Integer.toString(sourceFileReader_.getLineNumber()) +
                             ".");
      }
      // Set the line feature "red_id" attribute.
      feature.setIntAttribute("red_id",idNum);
      // Return the line feature.
      return feature;
   }

   // -----------------------------------------------------------------------
   // readTextFeature
   //
   // Read in a text feature. For text source files, each source line represents one feature.
   private IFMEFeature readTextFeature(String sourceLine, IFMEFeature feature) throws Exception
   {
      // Get tokenizer on the source line.
      StringTokenizer sourceLineTokenizer = new StringTokenizer(sourceLine, kComma);

      int idNum = 0;

      double xCoord = 0.0,
             yCoord = 0.0,
             zCoord = 0.0,
             rotationAngle = 0.0,
             textSize = 0.0;

      String textString = null;;

      try
      {
         // Extract the id, x & y coords, text angle, size and string from the line.
         idNum = Integer.parseInt(sourceLineTokenizer.nextToken());
         xCoord = Double.valueOf(sourceLineTokenizer.nextToken()).doubleValue();
         yCoord = Double.valueOf(sourceLineTokenizer.nextToken()).doubleValue();
         rotationAngle = Double.valueOf(sourceLineTokenizer.nextToken()).doubleValue();
         textSize = Double.valueOf(sourceLineTokenizer.nextToken()).doubleValue();
         textString = sourceLineTokenizer.nextToken();
      }
      catch (NoSuchElementException e)
      {
         // Throw exception if there aren't enough tokens in the source line.
         throw new Exception("Not enough values in text source file " +
                             sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                             " at line " +
                             Integer.toString(sourceFileReader_.getLineNumber()) + ".");
      }
      catch (NumberFormatException e)
      {
         // Throw exception if there are non-numeric values in the source line.
         throw new Exception("Non-numeric value for feature in text source file " +
                             sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                            " at line " +
                            Integer.toString(sourceFileReader_.getLineNumber()) +
                            ".");
      }
      // Add the coords to the text feature.
      feature.add3DCoordinate(xCoord, yCoord, zCoord);
      // Set the text feature attributes.
      feature.setIntAttribute("red_id",idNum);
      feature.setDoubleAttribute("red_rotation",rotationAngle);
      feature.setDoubleAttribute("red_text_size",textSize);
      feature.setStringAttribute("red_text_string",textString);
      // Return the text feature.
      return feature;
   }

   // -----------------------------------------------------------------------
   // readSchema
   //
   // Return the next schema feature for the data source, filling in the
   // empty feature object provided.  If an error occurs, an exception
   // is thrown, otherwise a returned feature object indicates that all
   // is well.
   public IFMEFeature readSchema(IFMEFeature schemaFeature) throws Exception
   {
//      logFile_.logMessageString("In readSchema() ******",IFMELogFile.FME_INFORM);
      if (nextSourceFileOpen_)
      {
         // Set the feature type (same as cource file's basename).
         schemaFeature.setFeatureType(currentSourceBasename_);
         // Set the "red_id" attribute common to all geometry types.
         schemaFeature.setStringAttribute("red_id","number(15,0)");

         if (currentSourceGeomType_.equals("red_line"))
         {
            // Set geometry for line type.
            schemaFeature.setStringAttribute("fme_geometry{0}", "red_line");
         }
         else if (currentSourceGeomType_.equals("red_point"))
         {
            // Set geometry for point type.
            schemaFeature.setStringAttribute("fme_geometry{0}", "red_point");
         }
         else if (currentSourceGeomType_.equals("red_text"))
         {
            // Set geometry and geometry attributes for text type.
            schemaFeature.setStringAttribute("fme_geometry{0}", "red_text");
            schemaFeature.setStringAttribute("fme_geomattr{0}.red_rotation", "number(9,4)");
            schemaFeature.setStringAttribute("fme_geomattr{0}.red_text_size", "number(15,4)");
            schemaFeature.setStringAttribute("fme_geomattr{0}.red_text_string", "char(80)");
         }
         else
         {
            // Otherwise throw exception for unknown source geometry.
            throw new Exception("Unknown source geometry type " +
                                currentSourceGeomType_ +
                                " for source feature type " +
                                currentSourceBasename_ + ".");
         }

         try
         {
            // Close the source file.
            sourceFileReader_.close();
         }
         catch (IOException e)
         {
            // Throw exception if problem closing.
            throw new Exception("Problem closing source file " +
                                sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                                ".");
         }
         // Open the next source file, saving return status for checking later if
         // another one is available.
         nextSourceFileOpen_ = openNextSourceFile();
         // Return the schema feature.
         return schemaFeature;
      }
      else
      {
         // Otherwise all source files have been processed so return null.
         return null;
      }
   }

   // -----------------------------------------------------------------------
   // spatialEnabled
   //
   //  Indicates whether or not this reader supports spatial contraints.
   //  By default this returns false.  If a reader does support
   //  spatial constraints then it returns true. */
   public boolean spatialEnabled() throws Exception
   {
      return false;
   }

   // -----------------------------------------------------------------------
   // setConstraints
   //
   //  Specifies the spatial and attribute constraints to be used
   //  when reading the data.  The spatial and attribute query
   //  is specified in the feature object. This only need be implemented
   //  by those readers which return spatialEnables() as true, otherwise
   //  they can do nothing.  This can be called at any time after the
   //  reader is created.  If any read is in progress then it is terminated
   //  and the next read will reflect the new constraints.
   //
   //  The constraint feature must have an attribute called 'fme_search_type'
   //  on it. The value of this attribute indicates the kind of search to
   //  be done.  Currently, the only supported values are:
   //
   //  fme_envelope_intersects : When this is the value, then the bounding
   //  box of the feature is used as the query envelope.  Future reads
   //  of the data will return only those features in the envelope area.
   //
   //  fme_closest : When this is the value, then a future read will
   //  only return the single feature that was the closest to the center of
   //  the bounding box of the passed in feature.  Normally, a point feature
   //  is used.  In addition, when this search method is used, a maximum
   //  distance must be given in an additional 'fme_max_distance' attribute.
   //
   //  Constants are provided at the top of this header file for these
   //  attribute names and values.  */
   public void setConstraints(IFMEFeature feature) throws Exception
   {
   }

   // -----------------------------------------------------------------------
   // openNextSourceFile
   //
   // Open the next source file in the source file array.
   private boolean openNextSourceFile() throws Exception
   {
      if (currentSourceFileNum_ < (numSourceFiles_ - 1))
      {
         // If there are still source files to process, construct next source file's
         // basename.
         ++currentSourceFileNum_;
         String sourceFilename = sourceFiles_[currentSourceFileNum_].getName();
         currentSourceBasename_ = sourceFilename.substring(0,sourceFilename.indexOf("."));

         try
         {
            // Get a line reader on the source file. The line reader is useful for text-based
            // formats like Red since it can report the current file line number, which can be
            // used in error messages.
            sourceFileReader_ = new LineNumberReader(new FileReader(sourceFiles_[currentSourceFileNum_]));
         }
         catch (FileNotFoundException e)
         {
            // Throw exception if source file not found.
            throw new Exception("Cannot find source file " +
                                sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                                ".");
         }
         // Get the geometry type for this source file.
         currentSourceGeomType_ = getGeomType();

         if (currentSourceGeomType_ == "red_text")
         {
            // If it's a text file, set the dimensionality to 2D.
            //          feature_.setDimension(IFMEFeature.FME_TWO_D);
         }
         // This source file is newly opened, so flag it to have its dimensionality checked.
         checkDimension_ = true;
         // Return true to indicate the source file is open and available for reading.
         return true;
      }
      else
      {
         // Otherwise all source files have been processed so return false.
         checkDimension_ = false;
         return false;
      }
   }

   // -----------------------------------------------------------------------
   // getGeomType
   //
   // Determine the geometry type of the current, newly-opened source file.
   private String getGeomType() throws Exception
   {
      String sourceLine;

      try
      {
         // Mark the read position at the first line, read the first line in and reset the
         // read position back to the first line. This allows the method to peek at
         // the line but still restore the read position to the first line for
         // subsequent read operations.
         sourceFileReader_.mark(kLookAheadLimit);
         sourceLine = sourceFileReader_.readLine();
         sourceFileReader_.reset();
      }
      catch (IOException e)
      {
         // Throw exception if problem reading the first source line.
         throw new Exception( "Problem reading source file " +
                              sourceFiles_[currentSourceFileNum_].getAbsolutePath() +
                              " at line " +
                              Integer.toString(sourceFileReader_.getLineNumber()) + ".");
      }
      // Get tokenizer on the first source line.
      StringTokenizer sourceLineTokenizer = new StringTokenizer(sourceLine, kComma);

      // Get the number of tokens on the first source line.
      int numTokens = sourceLineTokenizer.countTokens();

      if (numTokens == 1)
      {
         // If there's only one token (the id) it's a line.
         return "red_line";
      }
      else if ((numTokens == 3) || (numTokens == 4))
      {
         // If there are 3 or 4 tokens (the id and x,y and maybe z coords) it's a point.
         return "red_point";
      }
      else if (numTokens == 6)
      {
         // If there are 6 tokens (the id, x & y coords and the text attributes) it's text.
         return "red_text";
      }
      else
      {
         // Otherwise throw exception for unknown source geometry.
         throw new Exception( "Unknown source geometry type " +
                              currentSourceGeomType_ +
                              " for source feature type " +
                              currentSourceBasename_ + ".");
      }
   }

   // -----------------------------------------------------------------------
   // accept
   //
   // Method for returning source files with the correct extension for the Red format.
   // This is the implementation of the accept() method required for this
   // class to implement the FilenameFilter interface. Files accepted by this method
   // end up in the sourceFiles_ array for processing.
   public boolean accept(File dir, String name)
   {
      if (name.endsWith("." + kFormatExtension))
      {
         // If the file currently under consideration ends with the required extension,
         // accept it.
         return true;
      }
      else
      {
         // Otherwise reject it.
         return false;
      }
   }
}


