/*============================================================================

   Name     : RedWriter.java

   System   : FME Developer Kit

   Language : Java

   Purpose  : Implements the IFMEReader interface to supply a reader for
              the Red format (near-identical to Arc/Info Generate).

   Author               Date            Changes made
   ------------------   ------------    -------------------------------
   Your Name            Xxx XX, 200X    Original definition

                  Copyright (c) 1999 - 2000, Your Company Goes Here
                              All Rights Reserved

    The entire risk as to the results and performance of the software,
    supporting text and other information contained in this file
    (collectively called the "Software") is with the user.  Although
    Safe Software Incorporated has used considerable efforts in preparing
    the Software, Safe Software Incorporated does not warrant the
    accuracy or completeness of the Software. In no event will Safe Software
    Incorporated be liable for damages, including loss of profits or
    consequential damages, arising out of the use of the Software.

    NOTE: Replace the above header with whatever is appropriate for your
    company.

============================================================================*/

package com.mycompany.myformat.mywriter;

import COM.safe.fme.pluginbuilder.*;
import COM.safe.fmeobjects.*;
import java.io.*;
import java.util.*;


/******************************************************************************
Class Overview:
================
Implements the IFMEWriter interface to supply a writer for the Red format.

With three exceptions, the Red Writer is functionally identical to the FME ARCGEN Writer
used to write ESRI's Arc/Info Generate format. The two exceptions are:

1) Red does not require nor use any DEF lines in the mapping file.
2) Only the comma character is valid as the delimiter character in its destination files.
3) The data file extension must be ".red".

Details of the ARCGEN Writer (and therefore of the Red Writer) and the Generate
format can be found in the FME Reference Manual Part 2, in the Section
"ESRI ArcInfo Generate File Reader/Writer".
******************************************************************************/

//---------------------------------------------------------------------
// RedWriter class.
public final class RedWriter implements IFMEWriter
{
   // Public constants.
   public static final String kPoint = "_point";
   public static final String kLine = "_line";
   public static final String kText = "_text";
   public static final String kRedID = "red_id";
   public static final String kRedRotation = "red_rotation";
   public static final String kRedSize = "red_text_size";
   public static final String kRedText = "red_text_string";
   public static final String kComma = ",";
   public static final String kEND = "END";
   public static final int kDefaultNumCoords = 100;

   // The extension for Red source files.
   public static final String kFormatExtension = "red";

   private int currentSourceDimension_ = IFMEFeature.FME_TWO_D;

   // Private instance variables.
   private IFMEMappingFile mappingFile_ = null;
   private IFMELogFile logFile_ = null;
   private IFMECoordSysManager coordSysMan_ = null;

   // The Red exception message resource basename and locale.
   private String exceptionMsgResourceBasename_ = "redExceptions";
   private Locale exceptionMsgLocale_ = new Locale("en", "");

   private String writerType_ = null;
   private String writerKeyword_ = null;

   private String currentDestBasename_ = null;
   private String currentDestGeomType_ = null;

   private String destDataset_ = null;

   private String fileName_ = "";
   private FileWriter outputFile_ = null;

   // -----------------------------------------------------------------------
   // Public constructor.
   //
   // Takes the writer type and the writer keyword.
   public RedWriter(String writerType, String writerKeyword)
   {
      writerType_ = writerType;
      writerKeyword_ = writerKeyword;
      System.out.println("Writer Constructor: " + writerKeyword_ + " & " + writerType_);
   }

   // -----------------------------------------------------------------------
   // init
   //
   // Initialize the writer with the mapping file, log file, coodinate system
   // manager and session objects.
   //
   public void init(IFMEMappingFile mappingFile,
               IFMELogFile logFile,
               IFMECoordSysManager coordSysMan,
               IFMESession session)
               throws Exception
   {
      mappingFile_ = mappingFile;
      logFile_     = logFile;
      coordSysMan_ = coordSysMan;


/*
    ********************************* NOTE **************************************
    The mapping and log files are handed in to this method to enable the Writer to
    remember what they are for subsequent non-initialization usage.

    However, the mapping file should not be accessed from within this init() method,
    since there is no a priori way of determining whether or not it has usable content
    at this point. Mapping file usage should occur during the first and subsequent read()
    operations instead.
    *****************************************************************************

*/
   }

   public void configure() throws Exception
   {
      // do nothing
   }

   // -----------------------------------------------------------------------
   // open
   //
   // Open up the writer and pass it the destination directory to which files will
   // be written (1st parameter) and a string array which may contain additional
   // dataset parameters (2nd parameter). For example, the string array (the 2nd
   // parameter) may include one or more occurrences of the pair
   // "+ID" followed by the basename of a Red destination file.
   //
   // If such "+ID" pairs are present, the writer will write only those features that
   // correspond to these files, ignoring all other features passed to the writer.
   // If no "+ID" pairs are present, the writer writes out all the features
   // it is given into the appropriate files within the destination directory.
   //
   // For example, datasetParams could contain the following 4 elements:
   //
   //                "+ID" "road" "+ID" "sewer"
   //
   // The first string is a "+ID" keyword followed by the basename "road". Another
   // "+ID"-"basename" pair follows this, specifying a basename of "sewer". This
   // indicates to the Writer that it should only write (the corresponding) Red features
   // to "road_xxxx.red" and "sewer_xxxx.red" where "xxxx" is either "point", "line",
   // or "text". The two files "road.red" and "sewer.red" would be located
   // in the destination dataset directory indicated by the first parameter.
   //
   // If anything goes wrong an exception is thrown describing the type of error that occurred.
   public void open(ArrayList datasetParams) throws Exception
   {

      destDataset_ = (String)(datasetParams.get(0));

      // Make sure that at the end of the destination directory there
      // is a slash (i.e. '/' or '\\')
      if (destDataset_.charAt(destDataset_.length() -1) != '\\' ||
                   destDataset_.charAt(destDataset_.length() -1) != '/')
      {
         destDataset_ = destDataset_ + "/";
      }
   }

   // -----------------------------------------------------------------------
   // abort
   //
   // Terminate the writer mid-stream.  Any special actions to shut
   // down a writer not finished writing data should be taken in this
   // method.  For most writers, nothing will be done.
   public void abort() throws Exception
   {
      // do nothing
   }

   // -----------------------------------------------------------------------
   // close
   //
   // Close the writer after it has completed writing out all the features.
   public void close() throws Exception
   {
      if (outputFile_ != null)
      {
         try
         {
            outputFile_.write(kEND + "\n");
            outputFile_.close();
         }
         catch (IOException ioError)
         {
            throw new Exception("Problem closing destination file " + fileName_ + ".");
         }
      }
   }

   // -----------------------------------------------------------------------
   // id
   //
   // Return the unique ID assigned to this writer.
   public int id()
   {
      // All Java writers return 0.
      return 0;
   }

   // -----------------------------------------------------------------------
   // write
   //
   // Write the next feature to a file. If an error occurs, an exception
   // is thrown.
   public void write(IFMEFeature feature) throws Exception
   {
      String featureType;
      int fmeGeometry;
      boolean containsText;
      int indexToFileVector;

      featureType = feature.getFeatureType();
      fmeGeometry = feature.getGeometryType();
      containsText = isTextFeature(feature);

      if (!fileCorrespondingToFeatureIsOpen(featureType, fmeGeometry, containsText))
      {
         String stringFmeGeometry;
         switch (fmeGeometry)
         {
            case IFMEFeature.FME_GEOM_UNDEFINED:
               stringFmeGeometry = "UNDEFINED";
               break;
            case IFMEFeature.FME_GEOM_POINT:
               stringFmeGeometry = "POINT";
               break;
            case IFMEFeature.FME_GEOM_LINE:
               stringFmeGeometry = "LINE";
               break;
            case IFMEFeature.FME_GEOM_POLYGON:
               stringFmeGeometry = "POLYGON";
               break;
            case IFMEFeature.FME_GEOM_DONUT:
               stringFmeGeometry = "DONUT";
               break;
            case IFMEFeature.FME_GEOM_PIP:
               stringFmeGeometry = "PIP";
               break;
            case IFMEFeature.FME_GEOM_AGGREGATE:
               stringFmeGeometry = "AGGREGATE";
               break;
            default:
               stringFmeGeometry = "UNKNOWN";
               break;
         }
         createFeatureFile(featureType, fmeGeometry, containsText);
      }

      if (containsText)
      {
         addText(feature);
      }
      else if (fmeGeometry == IFMEFeature.FME_GEOM_POINT)
      {
         addPoint(feature);
      }
      else if (feature != null)
      {
         addLine(feature);
      }
  }

   //------------------------------------------------------------------------
   // isTextFeature
   //
   private boolean isTextFeature(IFMEFeature feat)
   {
      return feat.attributeExists(kRedText);
   }

   //------------------------------------------------------------------------
   // fileCorrespondingToFeatureIsOpen
   //
   private boolean fileCorrespondingToFeatureIsOpen(String featType,
                                                    int fmeGeom,
                                                    boolean hasText) throws Exception
   {
      if (fileName_.equals(""))
      {
         return false;
      }
      else
      {
         return true;
      }
   }

   //------------------------------------------------------------------------
   // createFeatureFile
   //
   private void createFeatureFile(String featType, int fmeGeometry, boolean hasText)
                                                                     throws Exception
   {
      if (hasText)
      {
         fileName_ = destDataset_ + featType + kText;
      }
      else if (fmeGeometry == IFMEFeature.FME_GEOM_POINT)
      {
         fileName_ = destDataset_ + featType + kPoint;
      }
      else
      {
         fileName_ = destDataset_ + featType + kLine;
      }

      fileName_ = fileName_ + "." + kFormatExtension;

      try
      {
         outputFile_ = new FileWriter(fileName_);
      }
      catch(IOException ioError)
      {
         throw new Exception("Was unable to open/create file " + fileName_ + " for output.");
      }
   }

   //------------------------------------------------------------------------
   // addPoint
   //
   private void addPoint(IFMEFeature feat) throws Exception
   {
      String idNumber = feat.getStringAttribute(kRedID);
      int dimension = feat.getDimension();
      double[] coords = new double[dimension];

      if (dimension == IFMEFeature.FME_TWO_D)
      {
         // Points contains only one coordinate
         coords = feat.get2DCoordinate(0);
      }
      else
      {
         // Points contains only one coordinate
         coords = feat.get3DCoordinate(0);
      }

      String xCoord = String.valueOf(coords[0]);
      String yCoord = String.valueOf(coords[1]);

      String outputString = idNumber + kComma + xCoord + kComma + yCoord;

      if (dimension == IFMEFeature.FME_THREE_D)
      {
         String zCoord = String.valueOf(coords[2]);
         outputString = outputString + kComma + zCoord;
      }

      outputString = outputString + "\n";

      try
      {
         outputFile_.write(outputString);
      }
      catch (IOException ioError)
      {
         throw new Exception("Unable to write the line `" +
                             outputString +
                             "' to file " +
                             fileName_ +
                             ".");
      }

   }

   //------------------------------------------------------------------------
   // addLine
   //
   private void addLine(IFMEFeature feat) throws Exception
   {
      String outputString;
      String idNumber = feat.getStringAttribute(kRedID);
      outputString = idNumber + "\n";

      int dimension = feat.getDimension();
      int numCoords = feat.numCoords();
      double[] coords = new double[dimension];

      for (int i=0; i<numCoords; i++)
      {
         if (dimension == IFMEFeature.FME_TWO_D)
         {
            coords = feat.get2DCoordinate(i);
         }
         else
         {
            coords = feat.get3DCoordinate(i);
         }

         outputString = outputString + (String.valueOf(coords[0]) + kComma + String.valueOf(coords[1]));
         if (dimension == IFMEFeature.FME_THREE_D)
         {
            outputString = outputString + (kComma + String.valueOf(coords[2]));
         }
         outputString = outputString + ("\n");
      }

      outputString = outputString + (kEND + "\n");

      try
      {
         outputFile_.write(outputString);
      }
      catch(IOException ioError)
      {
         throw new Exception("Unable to write the line `" +
                             outputString +
                             "' to file " +
                             fileName_ +
                             ".");
      }
   }

   //------------------------------------------------------------------------
   // addText
   //
   private void addText(IFMEFeature feat) throws Exception
   {
      String idNumber = feat.getStringAttribute(kRedID);

      double[] coords = new double[2];
      coords = feat.get2DCoordinate(0);     // Text features contain only one coordinate

      String xCoord = String.valueOf(coords[0]);
      String yCoord = String.valueOf(coords[1]);

      String outputString = idNumber + kComma + xCoord + kComma + yCoord;

      String rotation = feat.getStringAttribute(kRedRotation);
      String size = feat.getStringAttribute(kRedSize);
      String text = feat.getStringAttribute(kRedText);

      outputString = outputString + (kComma + rotation + kComma + size + kComma + text + "\n");

      try
      {
         outputFile_.write(outputString);
      }
      catch(IOException ioError)
      {
         throw new Exception("Unable to write the line `" +
                             outputString +
                             "' to file " +
                             fileName_ +
                             ".");
      }
   }

   //------------------------------------------------------------------------
   // addGlobalMetaData
   //
   public void addGlobalMetaData(IFMEFeature metaData) throws Exception
   {
      // do nothing
   }

   //------------------------------------------------------------------------
   // multiFileWriter
   //
   public boolean multiFileWriter()
   {
      return true;
   }

   //------------------------------------------------------------------------
   // startTransaction
   //
   public void startTransaction() throws Exception
   {
      // do nothing
   }

   //------------------------------------------------------------------------
   // commitTransaction
   //
   public void commitTransaction() throws Exception
   {
      // do nothing
   }

   //------------------------------------------------------------------------
   // rollbackTransaction
   //
   public void rollbackTransaction() throws Exception
   {
      // do nothing
   }
}
